program FMMusic;
{****************************************************************************
** Demonstration of directly programming the FM synthesizer on AdLib and   **
** Sound Blaster compatible cards                                          **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.home.ml.org                                               **
**                                                                         **
****************************************************************************}

uses Crt;

const {List of FNumbers}
  FNr : Array [0..11] of Word = (
  {C}     343,
  {C#/Db} 363,
  {D}     385,
  {D#/Eb} 408,
  {E}     432,
  {F}     458,
  {F#/Gb} 485,
  {G}     514,
  {G#/Ab} 544,
  {A}     577,
  {A#/Bb} 611,
  {B}     647);

  {Flat       Whole   Sharp}
              C =  0; CSharp =  1;
  DFlat =  1; D =  2; DSharp =  3;
  EFlat =  3; E =  4;
              F =  5; FSharp =  6;
  GFlat =  6; G =  7; GSharp =  8;
  AFlat =  8; A =  9; ASharp = 10;
  BFlat = 10; B = 11;


type Operator = record
  {Tremolo}
  Tremolo,
  {Vibrato}
  Vibrato,
  {Envelop generator type}
  EGType,
  {KSR}
  KSR,
  {Frequency multiplier}
  Multi,
  {Key scaling level}
  KSL,
  {Volume}
  TotalLevel,
  {Shape of the wave envelop}
  Attack, Decay, Sustain, Release,
  {Type of wave}
  WaveShape : Byte
end;

type Instrument = record
  Modulator,
  Carrier : Operator;
  {Feedback strength}
  Feedback,
  {Synthesis type}
  SynType : Byte;
end;

var
  Flute, Piano, Harp : Instrument;
  CurNote, CurBlock : Byte;

{**************************************************************************
** Writes a value to a specified index register on the FM card           **
**************************************************************************}
procedure WriteFM (Register, Value : Byte);
var
  Counter, Temp : Byte;

begin
  {Select register}
  Port [$388] := Register;
  {Wait for card to accept value}
  for Counter := 1 to 25 do Temp := Port [$388];
  {Send value}
  Port [$389] := Value;
  {Wait for card to accept value}
  for Counter := 1 to 100 do Temp := Port [$388];
end;

{**************************************************************************
** Checks for the presence of an FM card                                 **
**************************************************************************}
function FMInstalled : Boolean;
var
  A, B : Byte;

begin
  WriteFM (1,   0);
  WriteFM (4, $60);
  WriteFM (4, $80);
  A := Port [$388];
  WriteFM (2, $FF);
  WriteFM (4, $21);
  B := Port [$388];
  WriteFM (4, $60);
  WriteFM (4, $80);
  if ((A and $E0)=0) and ((B and $E0)=$C0) then
    FMInstalled:=true
  else
    FMInstalled:=false;
end;

{**************************************************************************
** Activates a voice on the FM card                                      **
***************************************************************************
** Voice selects one of the 9 FM voices                                  **
** FNumber selects the note to be played                                 **
** Block selects the octave for the specified note                       **
**************************************************************************}
procedure NoteOn (Voice, Note, Block : Byte);
begin
  WriteFM ($A0+Voice, lo(FNr[Note]));
  WriteFM ($B0+Voice, hi(FNr[Note]) + (Block shl 2) + 32);
end;

{**************************************************************************
** Deactivates a voice on the FM card                                    **
***************************************************************************
** Make sure to give the same values for Note and Block or this will     **
** sound very odd.                                                       **
**************************************************************************}
procedure NoteOff (Voice, Note, Block : Byte);
begin
  WriteFM ($A0+Voice, lo(FNr[Note]));
  WriteFM ($B0+Voice, hi(FNr[Note]) + (Block shl 2));
end;

{**************************************************************************
** Sets instrument settings for a voice on the FM card                   **
**************************************************************************}
procedure SetInstrument (Voice : Byte; Instr : Instrument);
const
  {Addresses of the operators used to form voice data}
  OpAdr : Array [0..8] of Byte = (0, 1, 2, 8, 9, 10, 16, 17, 18);

var
  Value : Byte;

begin
  {Set up voice modulator}
  Value := Instr.Modulator.Tremolo shl 8
        or Instr.Modulator.Vibrato shl 7
        or Instr.Modulator.EGType  shl 6
        or Instr.Modulator.KSR     shl 5
        or Instr.Modulator.Multi;
  WriteFM ($20 + OpAdr [Voice], Value);

  Value := Instr.Modulator.KSL shl 7
        or Instr.Modulator.TotalLevel;
  WriteFM ($40 + OpAdr [Voice], Value);

  Value := Instr.Modulator.Attack shl 4
        or Instr.Modulator.Decay;
  WriteFM ($60 + OpAdr [Voice], Value);

  Value := Instr.Modulator.Sustain shl 4
        or Instr.Modulator.Release;
  WriteFM ($80 + OpAdr [Voice], Value);

  WriteFM ($E0 + OpAdr [Voice], Instr.Modulator.WaveShape);

  Value := Instr.Feedback shl 1
        or Instr.SynType;
  WriteFM ($C0 + OpAdr [Voice], Value);

  {Set up voice carrier}
  Value := Instr.Carrier.Tremolo shl 8
        or Instr.Carrier.Vibrato shl 7
        or Instr.Carrier.EGType  shl 6
        or Instr.Carrier.KSR     shl 5
        or Instr.Carrier.Multi;
  WriteFM ($23 + OpAdr [Voice], Value);

  Value := Instr.Carrier.KSL shl 7
        or Instr.Carrier.TotalLevel;
  WriteFM ($43 + OpAdr [Voice], Value);

  Value := Instr.Carrier.Attack shl 4
        or Instr.Carrier.Decay;
  WriteFM ($63 + OpAdr [Voice], Value);

  Value := Instr.Carrier.Sustain shl 4
        or Instr.Carrier.Release;
  WriteFM ($83 + OpAdr [Voice], Value);

  WriteFM ($E3 + OpAdr [Voice], Instr.Carrier.WaveShape);
end;

begin

  ClrScr;
  Writeln ('Demonstration of programming an FM sound cards [AdLib, SB and compatibles]');
  Writeln ('--------------------------------------------------------------------------');

  {Check to see if an FM sound card is present}
  if FMInstalled then begin
    Writeln ('AdLib/SB or compatible FM sound card found.');
  end else begin
    {If not, end the programme}
    Writeln ('AdLib/SB or compatible FM sound card not found.');
    Halt;
  end;

  {Set up voices 0 - 3 to a "harp" - note, only the values other than 0 are set}
  Harp.Modulator.Attack     := 15;
  Harp.Modulator.Decay      :=  5;
  Harp.Modulator.Sustain    :=  8;
  Harp.Modulator.Release    :=  5;
  Harp.Modulator.Multi      :=  2;
  Harp.Modulator.TotalLevel := 41;

  Harp.Carrier.Attack       := 15;
  Harp.Carrier.Decay        :=  2;
  Harp.Carrier.Release      :=  3;
  Harp.Carrier.Multi        :=  1;
  Harp.Carrier.KSL          :=  2;
  Harp.Carrier.TotalLevel   :=  3;

  SetInstrument (0, Harp);
  SetInstrument (1, Harp);
  SetInstrument (2, Harp);

  {Play C chord using Harp}
  Writeln ('Playing C chord.');

  NoteOn  (0, C, 4);
  NoteOn  (1, E, 4);
  NoteOn  (2, G, 4);
  Delay   (1000); {Wait 1 second}
  NoteOff (0, C, 4);
  NoteOff (1, E, 4);
  NoteOff (2, G, 4);
  {Wait a while}
  Delay (500);

  {Set up voice 3 to a "piano" - this doesn't really sound like a piano,
  but that's what they call it!}
  Piano.Modulator.Attack     := 15;
  Piano.Modulator.Decay      :=  1;
  Piano.Modulator.Sustain    := 10;
  Piano.Modulator.Release    :=  3;
  Piano.Modulator.Multi      :=  1;
  Piano.Modulator.KSL        :=  1;
  Piano.Modulator.TotalLevel := 16;

  Piano.Carrier.Attack       := 13;
  Piano.Carrier.Decay        :=  2;
  Piano.Carrier.Sustain      :=  8;
  Piano.Carrier.Release      :=  4;
  Piano.Carrier.Multi        :=  1;
  Piano.Carrier.TotalLevel   :=  0;
  Piano.Carrier.KSR          :=  1;

  Piano.Feedback             :=  3;

  SetInstrument (3, Piano);

  {Play all notes in octave 4 for .1 seconds using piano [voice 3]}
  Writeln ('Playing all notes in octave four.');
  CurBlock := 4;
  for CurNote := 0 to 11 do begin
    NoteOn  (3, CurNote, CurBlock);
    Delay   (100); {Wait .1 seconds}
    NoteOff (3, CurNote, CurBlock);
  end;

  {Wait a while}
  Delay (500);

  {Set up voice 4 to a "flute"}
  Flute.Modulator.Attack     :=  6;
  Flute.Modulator.Decay      := 14;
  Flute.Modulator.Sustain    :=  7;
  Flute.Modulator.Release    := 15;
  Flute.Modulator.Vibrato    :=  1;
  Flute.Modulator.Tremolo    :=  1;
  Flute.Modulator.EGType     :=  1;
  Flute.Modulator.KSL        :=  3;
  Flute.Modulator.TotalLevel := 44;

  Flute.Carrier.Attack       :=  6;
  Flute.Carrier.Decay        :=  5;
  Flute.Carrier.Sustain      := 13;
  Flute.Carrier.Release      := 10;
  Flute.Carrier.Vibrato      :=  1;
  Flute.Carrier.EGType       :=  1;
  Flute.Carrier.Multi        :=  1;
  Flute.Carrier.TotalLevel   :=  0;

  Flute.Feedback             :=  7;
  Flute.SynType              :=  3;

  SetInstrument (4, Flute);

  {Play short tune using Flute [voice 4] as primary, Harp [voice 0] as secondary voice}
  Writeln ('Playing short tune.');

  NoteOn  (4, C, 5); NoteOn  (0, C, 3);
  Delay   (250);
  NoteOff (4, C, 5);
  NoteOn  (4, E, 5);
  Delay   (250);
  NoteOff (4, E, 5); NoteOff (0, C, 3);

  NoteOn  (4, C, 5); NoteOn  (0, G, 3);
  Delay   (250);
  NoteOff (4, C, 5);
  NoteOn  (4, G, 4);
  Delay   (250);
  NoteOff (4, G, 4); NoteOff (0, G, 3);

  NoteOn  (4, E, 4); NoteOn  (0, C, 4);
  Delay   (250);
  NoteOff (4, E, 4);
  NoteOn  (4, G, 4);
  Delay   (250);
  NoteOff (4, G, 4); NoteOff (0, C, 4);

  NoteOn  (4, E, 4); NoteOn  (0, G, 3);
  Delay   (250);
  NoteOff (4, E, 4);
  NoteOn  (4, D, 4);
  Delay   (250);
  NoteOff (4, D, 4); NoteOff (0, G, 3);

  NoteOn  (4, C, 4); NoteOn  (0, C, 3);
  Delay   (750);
  NoteOff (4, C, 4); NoteOff (0, C, 3);
end.